#define GLFW_INCLUDE_NONE
#define GLFW_INCLUDE_VULKAN
#include <GLFW/glfw3.h>
#include <cstdlib>
#include <iostream>
#include <stdexcept>

#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>

#include "fmt/base.h"
#include "util/types.h"

namespace vk {
  using DebugUtilsMessengerUnique = vk::UniqueHandle<vk::DebugUtilsMessengerEXT, vk::DispatchLoaderDynamic>;
}

constexpr int WIDTH  = 800;
constexpr int HEIGHT = 600;

constexpr std::array<const char*, 1> validationLayers = { "VK_LAYER_KHRONOS_validation" };

#ifdef NDEBUG
constexpr bool enableValidationLayers = false;
#else
constexpr bool enableValidationLayers = true;
#endif

class VulkanApp {
 public:
  void run() {
    initWindow();
    initVulkan();
    mainLoop();
    cleanup();
  }

 private:
  GLFWwindow* mWindow;

  vk::UniqueInstance mInstance;

  vk::DebugUtilsMessengerUnique mDebugMessenger;
  vk::DispatchLoaderDynamic     mLoader;

  vk::PhysicalDevice mPhysicalDevice;

  struct QueueFamilyIndices {
    std::optional<u32> graphics_family;

    fn isComplete() -> bool { return graphics_family.has_value(); }
  };

  void initWindow() {
    glfwInit();

    glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);
    glfwWindowHint(GLFW_RESIZABLE, GLFW_FALSE);

    mWindow = glfwCreateWindow(WIDTH, HEIGHT, "Vulkan", nullptr, nullptr);
  }

  void initVulkan() {
    createInstance();
    setupDebugMessenger();
    pickPhysicalDevice();
  }

  void mainLoop() {
    while (!glfwWindowShouldClose(mWindow)) { glfwPollEvents(); }
  }

  void cleanup() {
    glfwDestroyWindow(mWindow);
    glfwTerminate();
  }

  fn createInstance() -> void {
    if (enableValidationLayers && !checkValidationLayerSupport())
      throw std::runtime_error("validation layers requested, but not available!");

    vk::ApplicationInfo appInfo { .pApplicationName   = "Hello Triangle",
                                  .applicationVersion = 1,
                                  .pEngineName        = "No Engine",
                                  .engineVersion      = 1,
                                  .apiVersion         = VK_API_VERSION_1_0 };

    // Retrieve extensions using custom function
    std::vector<const char*> extensions = getRequiredExtensions();

    // Enable the portability extension and set flags
    extensions.push_back("VK_KHR_portability_enumeration");

    vk::InstanceCreateInfo createInfo {
      .flags                   = vk::InstanceCreateFlagBits::eEnumeratePortabilityKHR,
      .pApplicationInfo        = &appInfo,
      .enabledLayerCount       = enableValidationLayers ? static_cast<uint32_t>(validationLayers.size()) : 0,
      .ppEnabledLayerNames     = enableValidationLayers ? validationLayers.data() : nullptr,
      .enabledExtensionCount   = static_cast<uint32_t>(extensions.size()),
      .ppEnabledExtensionNames = extensions.data()
    };

#ifndef NDEBUG
    fmt::println("Available extensions:");

    for (const char* extension : extensions) fmt::println("\t{}", extension);
#endif

    try {
      mInstance = vk::createInstanceUnique(createInfo);
      mLoader   = vk::DispatchLoaderDynamic(mInstance.get(), vkGetInstanceProcAddr);
    } catch (const vk::SystemError& err) {
      throw std::runtime_error("Failed to create instance: " + std::string(err.what()));
    }
  }

  void setupDebugMessenger() {
    if (!enableValidationLayers)
      return;

    vk::DebugUtilsMessengerCreateInfoEXT messengerCreateInfo {
      .messageSeverity = vk::DebugUtilsMessageSeverityFlagBitsEXT::eVerbose |
                         vk::DebugUtilsMessageSeverityFlagBitsEXT::eWarning |
                         vk::DebugUtilsMessageSeverityFlagBitsEXT::eError,
      .messageType = vk::DebugUtilsMessageTypeFlagBitsEXT::eGeneral |
                     vk::DebugUtilsMessageTypeFlagBitsEXT::eValidation |
                     vk::DebugUtilsMessageTypeFlagBitsEXT::ePerformance,
      .pfnUserCallback = debugCallback,
    };

    mDebugMessenger = mInstance->createDebugUtilsMessengerEXTUnique(messengerCreateInfo, nullptr, mLoader);
  }

  fn pickPhysicalDevice() -> void {
    std::vector<vk::PhysicalDevice> devices = mInstance->enumeratePhysicalDevices();

    if (devices.empty())
      throw std::runtime_error("Failed to find GPUs with Vulkan support!");

#ifndef NDEBUG
    fmt::println("Available devices:");
#endif

    for (const auto& device : devices) {
#ifndef NDEBUG
      vk::PhysicalDeviceProperties properties = device.getProperties();
      fmt::println("Device: {}", properties.deviceName.data());
#endif

      if (isDeviceSuitable(device)) {
        mPhysicalDevice = device;
        break;
      }
    }

    if (!mPhysicalDevice)
      throw std::runtime_error("Failed to find a suitable GPU!");
  }

  static fn isDeviceSuitable(vk::PhysicalDevice device) -> bool {
    QueueFamilyIndices indices = findQueueFamilies(device);

    return indices.isComplete();
  }

  static fn findQueueFamilies(vk::PhysicalDevice device) -> QueueFamilyIndices {
    QueueFamilyIndices indices;

    std::vector<vk::QueueFamilyProperties> queueFamilies = device.getQueueFamilyProperties();

    for (u32 i = 0; i < queueFamilies.size(); i++) {
      if (queueFamilies[i].queueFlags & vk::QueueFlagBits::eGraphics)
        indices.graphics_family = i;

      if (indices.isComplete())
        break;
    }

    return indices;
  }

  static fn getRequiredExtensions() -> std::vector<const char*> {
    u32          glfwExtensionCount = 0;
    const char** glfwExtensions     = nullptr;
    glfwExtensions                  = glfwGetRequiredInstanceExtensions(&glfwExtensionCount);

    std::vector<const char*> extensions;

    if (glfwExtensions) {
      std::span<const char*> extSpan(glfwExtensions, glfwExtensionCount);
      extensions.assign(extSpan.begin(), extSpan.end());
    }

    if (enableValidationLayers)
      extensions.push_back("VK_EXT_debug_utils");

    return extensions;
  }

  static fn checkValidationLayerSupport() -> bool {
    std::vector<vk::LayerProperties> availableLayers = vk::enumerateInstanceLayerProperties();

    for (const char* layerName : validationLayers) {
      bool layerFound = false;

      for (const auto& layerProperties : availableLayers)
        if (strcmp(layerName, layerProperties.layerName) == 0) {
          layerFound = true;
          break;
        }

      if (!layerFound)
        return false;
    }

    return true;
  }

  static VKAPI_ATTR fn VKAPI_CALL debugCallback(
    VkDebugUtilsMessageSeverityFlagBitsEXT /*messageSeverity*/,
    VkDebugUtilsMessageTypeFlagsEXT /*messageType*/,
    const VkDebugUtilsMessengerCallbackDataEXT* pCallbackData,
    void* /*pUserData*/
  ) -> VkBool32 {
    fmt::println("Validation layer: {}", pCallbackData->pMessage);

    return VK_FALSE;
  }
};

fn main() -> i32 {
  VulkanApp app;

  try {
    app.run();
  } catch (const std::exception& e) {
    std::cerr << e.what() << std::endl;
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
